#ifndef F3DConfig_h
#define F3DConfig_h

#include <string>

namespace F3D
{
const std::string AppName = "F3D";
const std::string AppTitle = "F3D - A fast and minimalist 3D viewer";
const std::string AppVersion = "3.5.0";
const std::string AppVersionFull = "3.5.0";
const std::string PluginsInstallDir = "lib";
}
#endif
