// clang-format off
#if 1
bool canRead([[maybe_unused]] vtkResourceStream* stream) const override
{
#if 0
  std::string hint;
  bool ret = vtkF3DImageImporter::CanReadFile(stream, hint);
  ret &= (hint == "png");
  return ret;
#else
  return true;
#endif
}
#endif

void applyCustomImporter(vtkImporter* importer, const std::string& vtkNotUsed(fileName),
  vtkResourceStream* vtkNotUsed(stream)) const override
{
  // Set hint
  vtkF3DImageImporter* imageImporter = vtkF3DImageImporter::SafeDownCast(importer);
  imageImporter->SetImageHint("png");
}
// clang-format on
