// SPDX-FileCopyrightText: Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
// SPDX-License-Identifier: BSD-3-Clause
#ifndef vtkF3DBitonicSortGlobalFlipCS_h
#define vtkF3DBitonicSortGlobalFlipCS_h

extern const char *vtkF3DBitonicSortGlobalFlipCS;

#endif
