// /!\ Autogenerated file - do not modify

#define F3D_PLUGIN_IS_STATIC 1

#include "plugin.h"
#include "plugin_export.h"

#include <memory>

#include "reader_VTKHDF.h"
#include "reader_ExodusII.h"
#include "reader_NetCDF.h"


static std::shared_ptr<f3d::plugin> gPlugin;

#if F3D_PLUGIN_IS_STATIC
f3d::plugin* init_plugin_static_hdf()
#else
extern "C" F3D_PLUGIN_EXPORT f3d::plugin* init_plugin()
#endif
{
  if (!gPlugin)
  {
    gPlugin = std::make_shared<f3d::plugin>(
      "hdf",
      "VTK HDF formats support",
      "1.0",
      std::vector<std::shared_ptr<f3d::reader>>{ std::make_shared<reader_VTKHDF>(),std::make_shared<reader_ExodusII>(),std::make_shared<reader_NetCDF>(), }
    );
  }

  return gPlugin.get();
}
