// /!\ Autogenerated file - do not modify

#define F3D_PLUGIN_IS_STATIC 1

#include "plugin.h"
#include "plugin_export.h"

#include <memory>

#include "reader_Draco.h"
#include "reader_GLTFDraco.h"
#include "reader_GLBDraco.h"


static std::shared_ptr<f3d::plugin> gPlugin;

#if F3D_PLUGIN_IS_STATIC
f3d::plugin* init_plugin_static_draco()
#else
extern "C" F3D_PLUGIN_EXPORT f3d::plugin* init_plugin()
#endif
{
  if (!gPlugin)
  {
    gPlugin = std::make_shared<f3d::plugin>(
      "draco",
      "Draco support (version 1.5.7)",
      "1.0",
      std::vector<std::shared_ptr<f3d::reader>>{ std::make_shared<reader_Draco>(),std::make_shared<reader_GLTFDraco>(),std::make_shared<reader_GLBDraco>(), }
    );
  }

  return gPlugin.get();
}
