#include "config.h"

namespace f3d::detail
{
const std::string LibVersion = "3.5.0";          // Version is synchronized with F3D
const std::string LibVersionFull = "3.5.0"; // Version is synchronized with F3D
const std::string LibBuildSystem = "Linux 64-bits";
const std::string LibBuildDate = "2026-04-26 07:01:38";
const std::string LibCompiler = "GNU 15.2.1";
}
