// SPDX-FileCopyrightText: Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
// SPDX-License-Identifier: BSD-3-Clause
#ifndef vtkF3DPointSplatVS_h
#define vtkF3DPointSplatVS_h

extern const char *vtkF3DPointSplatVS;

#endif
