// SPDX-FileCopyrightText: Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
// SPDX-License-Identifier: BSD-3-Clause
#ifndef vtkF3DBitonicSortLocalDisperseCS_h
#define vtkF3DBitonicSortLocalDisperseCS_h

extern const char *vtkF3DBitonicSortLocalDisperseCS;

#endif
