// clang-format off
#if 1
bool canRead(vtkResourceStream* stream) const override
{
  std::string hint;
  bool ret = vtkF3DAssimpImporter::CanReadFile(stream, hint);
  ret &= (hint == "3mf");
  return ret;
}
#endif

void applyCustomImporter(vtkImporter* importer, const std::string& vtkNotUsed(fileName),
  vtkResourceStream* vtkNotUsed(stream)) const override
{
  // Set hint
  vtkF3DAssimpImporter* assimpImporter = vtkF3DAssimpImporter::SafeDownCast(importer);
  assimpImporter->SetMemoryHint("3mf");
}
// clang-format on
